# =============================================================================
# Minet CLI Utils
# =============================================================================
#
# Miscellaneous helpers used by the CLI tools.
#
import csv
import sys
import json
import codecs
import argparse
from glob import iglob
from os.path import join
from collections import namedtuple
from tqdm import tqdm


def print_err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def die(msg):
    if not isinstance(msg, list):
        msg = [msg]

    for m in msg:
        print_err(m)

    sys.exit(1)


def safe_index(l, e):
    try:
        return l.index(e)
    except ValueError:
        return None


def custom_reader(f, target_header):

    reader = csv.reader(f)

    headers = next(reader, None)

    if headers[0].startswith(codecs.BOM_UTF8.decode()):
        headers[0] = headers[0][1:]

    if isinstance(target_header, tuple):
        HeaderPositions = namedtuple('HeaderPositions', target_header)
        position = HeaderPositions(**{t: safe_index(headers, t) for t in target_header})
    else:
        position = safe_index(headers, target_header)

    return headers, position, reader


class CSVEnricher(object):
    def __init__(self, f, target, output_file, report_headers, select=None):
        headers, pos, reader = custom_reader(f, target)

        self.report_headers = report_headers
        self.pos = pos
        self.reader = reader
        self.padding = [''] * len(report_headers)

        self.select = select
        self.selected_pos = None

        if select is not None:
            self.selected_pos = [headers.index(h) for h in select]
            headers = select

        writer = csv.writer(output_file)
        writer.writerow(headers + report_headers)
        self.writer = writer

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.reader)

    def __filter_line(self, line):
        if self.selected_pos is not None:
            return [line[p] for p in self.selected_pos]

        return line

    def write_empty(self, line):
        line = self.__filter_line(line)

        self.writer.writerow(line + self.padding)

    def write(self, line, row):
        assert len(row) == len(self.report_headers)

        line = self.__filter_line(line)

        self.writer.writerow(line + row)


def url_parse_action(namespace):

    headers, pos, reader = custom_reader(namespace.file, namespace.column)

    if namespace.select:
        selected_headers = namespace.select.split(',')
        selected_pos = [headers.index(h) for h in selected_headers]
        headers = selected_headers

    if namespace.output is None:
        output_file = DummyTqdmFile()
    else:
        output_file = open(namespace.output, 'w')

    writer = csv.writer(output_file)
    writer.writerow(headers + REPORT_HEADERS)


class DummyTqdmFile(object):
    """
    Dummy file-like that will write to tqdm. Taken straight from the lib's
    documentation: https://github.com/tqdm/tqdm but modified for minet use
    case regarding stdout piping.
    """
    file = None

    def __init__(self, file=sys.stdout):
        self.file = file

    def write(self, x):
        # Avoid print() second call (useless \n)
        tqdm.write(x, file=self.file, end='')

    def flush(self):
        return self.file.flush()

    def close(self):
        pass


def open_output_file(output, flag='w'):
    if output is None:
        return DummyTqdmFile(sys.stdout)

    return open(output, flag)


class BooleanAction(argparse.Action):
    """
    Custom argparse action to handle --no-* flags.
    Taken from: https://thisdataguy.com/2017/07/03/no-options-with-argparse-and-python/
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super(BooleanAction, self).__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, False if option_string.startswith('--no') else True)


class JSONLWriter(object):
    def __init__(self, file):
        self.file = file

    def writerow(self, item):
        row = json.dumps(item, ensure_ascii=False)
        self.file.write(row + '\n')


WorkerPayload = namedtuple(
    'WorkerPayload',
    ['line', 'headers', 'path', 'encoding', 'content', 'args']
)


def create_report_iterator(namespace, args=None, loading_bar=None):
    input_headers, pos, reader = custom_reader(namespace.report, ('status', 'filename', 'encoding', 'raw_content'))

    indexed_headers = {h: p for p, h in enumerate(input_headers)}

    for line in reader:
        status = int(line[pos.status]) if line[pos.status] else None
        filename = line[pos.filename]

        if status is None or status >= 400 or not filename:
            if loading_bar is not None:
                loading_bar.update()
            continue

        if pos.raw_content is not None:
            yield WorkerPayload(
                line=line,
                headers=indexed_headers,
                path=None,
                encoding=None,
                content=line[pos.raw_content],
                args=args
            )

            continue

        path = join(namespace.input_directory, filename)
        encoding = line[pos.encoding].strip() or 'utf-8'

        yield WorkerPayload(
            line=line,
            headers=indexed_headers,
            path=path,
            encoding=encoding,
            content=None,
            args=args
        )


def create_glob_iterator(namespace, args):
    for p in iglob(namespace.glob, recursive=True):
        yield WorkerPayload(
            line=None,
            headers=None,
            path=p,
            encoding='utf-8',
            content=None,
            args=args
        )


class LazyLineDict(object):
    def __init__(self, headers, line):
        self.headers = headers
        self.line = line

    def __getitem__(self, key):
        return self.line[self.headers[key]]



# =============================================================================
# Minet CrowdTangle Links Summary CLI Action
# =============================================================================
#
# Logic of the `ct summary` action.
#
import csv
from io import StringIO
from urllib.parse import quote
from tqdm import tqdm
from ural import is_url

from minet.utils import create_pool, request_json, RateLimiter, nested_get
from minet.cli.utils import die, custom_reader

from minet.cli.crowdtangle.constants import (
    CROWDTANTLE_LINKS_DEFAULT_RATE_LIMIT,
    CROWDTANGLE_REACTION_TYPES,
    CROWDTANGLE_DEFAULT_TIMEOUT
)
from minet.cli.crowdtangle.posts import (
    format_post_for_csv,
    CSV_HEADERS as POST_REPORT_CSV_HEADERS
)

POST_REPORT_CSV_HEADERS = ['url'] + POST_REPORT_CSV_HEADERS

URL_TEMPLATE = (
    'https://api.crowdtangle.com/links'
    '?token=%(token)s'
    '&count=%(count)s'
    '&startDate=%(start_date)s'
    '&includeSummary=true'
    '&link=%(link)s'
    '&sortBy=%(sort_by)s'
)


def forge_url(namespace, link):
    return URL_TEMPLATE % {
        'token': namespace.token,
        'count': 1 if namespace.posts is None else 100,
        'start_date': namespace.start_date,
        'link': quote(link, safe=''),
        'sort_by': namespace.sort_by
    }


CSV_HEADERS = ['%s_count' % t for t in CROWDTANGLE_REACTION_TYPES]
CSV_PADDING = ['0'] * len(CSV_HEADERS)


def format_summary_for_csv(stats):
    return [stats['%sCount' % t] for t in CROWDTANGLE_REACTION_TYPES]


def crowdtangle_summary_action(namespace, output_file):
    if not namespace.start_date:
        die('Missing --start-date!')

    http = create_pool(timeout=CROWDTANGLE_DEFAULT_TIMEOUT)

    if is_url(namespace.column):
        namespace.file = StringIO('url\n%s' % namespace.column)
        namespace.column = 'url'

    input_headers, pos, reader = custom_reader(namespace.file, namespace.column)

    output_headers = input_headers + CSV_HEADERS
    output_writer = csv.writer(output_file)
    output_writer.writerow(output_headers)

    posts_writer = None

    if namespace.posts is not None:
        posts_writer = csv.writer(namespace.posts)
        posts_writer.writerow(POST_REPORT_CSV_HEADERS)

    rate_limit = namespace.rate_limit if namespace.rate_limit is not None else CROWDTANTLE_LINKS_DEFAULT_RATE_LIMIT
    rate_limiter = RateLimiter(rate_limit, 60.0)

    loading_bar = tqdm(
        desc='Collecting data',
        dynamic_ncols=True,
        total=namespace.total,
        unit=' urls'
    )

    for line in reader:
        with rate_limiter:
            link = line[pos]

            # Fetching
            err, response, data = request_json(http, forge_url(namespace, link))

            # Handling errors
            if err:
                die(err)
            elif response.status == 401:
                die([
                    'Your API token is invalid.',
                    'Check that you indicated a valid one using the `--token` argument.'
                ])
            elif response.status >= 400:
                die(response.data, response.status)

            summary = nested_get(['result', 'summary', 'facebook'], data)

            if summary is None:
                output_writer.writerow(line + CSV_PADDING)
            else:
                output_writer.writerow(line + format_summary_for_csv(summary))

            posts = nested_get(['result', 'posts'], data)

            if posts is not None and namespace.posts is not None:
                for post in posts:
                    posts_writer.writerow([link] + format_post_for_csv(namespace, post))

            loading_bar.update()



# =============================================================================
# Minet CrowdTangle CLI Defaults
# =============================================================================
#
# General constants used throughout the CrowdTangle actions.
#
from urllib3 import Timeout

CROWDTANGLE_DEFAULT_RATE_LIMIT = 6  # Number of hits per minute
CROWDTANTLE_LINKS_DEFAULT_RATE_LIMIT = 2

CROWDTANGLE_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 5)

CROWDTANGLE_POST_TYPES = [
    'episode',
    'extra_clip',
    'link',
    'live_video',
    'live_video_complete',
    'live_video_scheduled',
    'native_video',
    'photo',
    'status',
    'trailer',
    'tweet',
    'video',
    'vine',
    'youtube'
]

CROWDTANGLE_SORT_TYPES = [
    'date',
    'interaction_rate',
    'overperforming',
    'total_interactions',
    'underperforming'
]

CROWDTANGLE_PARTITION_STRATEGIES = {
    'day'
}

CROWDTANGLE_PLATFORMS = {
    'facebook',
    'instagram',
    'reddit'
}

CROWDTANGLE_REACTION_TYPES = [
    'angry',
    'comment',
    'haha',
    'like',
    'love',
    'sad',
    'share',
    'thankful',
    'wow'
]



# =============================================================================
# Minet CrowdTangle Posts CLI Action
# =============================================================================
#
# Logic of the `ct posts` action.
#
import csv
import json

from minet.utils import create_pool, request
from minet.cli.utils import die

URL_TEMPLATE = 'https://api.crowdtangle.com/lists?token=%s'

CSV_HEADERS = [
    'id',
    'title',
    'type'
]


def format_list_for_csv(l):
    return [
        l['id'],
        l['title'],
        l['type']
    ]


def crowdtangle_lists_action(namespace, output_file):
    http = create_pool()

    _, result = request(http, URL_TEMPLATE % namespace.token)

    if result.status == 401:
        die([
            'Your API token is invalid.',
            'Check that you indicated a valid one using the `--token` argument.'
        ])

    if result.status >= 400:
        die([result.data, result.status])

    writer = csv.writer(output_file)
    writer.writerow(CSV_HEADERS)

    data = json.loads(result.data)

    for l in data['result']['lists']:
        writer.writerow(format_list_for_csv(l))



# =============================================================================
# Minet Extract Content CLI Action
# =============================================================================
#
# Logic of the extract action.
#
import csv
import gzip
import codecs
import warnings
from multiprocessing import Pool
from tqdm import tqdm
from dragnet import extract_content

from minet.encodings import is_supported_encoding
from minet.cli.utils import (
    custom_reader,
    open_output_file,
    create_report_iterator
)
from minet.cli.reporters import report_error

from minet.exceptions import UnknownEncodingError

OUTPUT_ADDITIONAL_HEADERS = ['extract_error', 'extracted_text']


def worker(payload):
    line, _, path, encoding, content, _ = payload

    if not is_supported_encoding(encoding):
        return UnknownEncodingError('Unknown encoding: "%s"' % encoding), line, None

    # Reading file
    try:
        if path.endswith('.gz'):
            with open(path, 'rb') as f:
                raw_html_bytes = gzip.decompress(f.read())

            raw_html = raw_html_bytes.decode(encoding, errors='replace')
        else:
            with codecs.open(path, 'r', encoding=encoding, errors='replace') as f:
                raw_html = f.read()
    except UnicodeDecodeError as e:
        return e, line, None

    # Attempting extraction
    try:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            content = extract_content(raw_html)
    except BaseException as e:
        return e, line, None

    return None, line, content


def extract_action(namespace):
    input_headers, pos, reader = custom_reader(namespace.report, ('status', 'filename', 'encoding'))

    selected_fields = namespace.select.split(',') if namespace.select else None
    selected_pos = [input_headers.index(h) for h in selected_fields] if selected_fields else None

    output_headers = (list(input_headers) if not selected_pos else [input_headers[i] for i in selected_pos])
    output_headers += OUTPUT_ADDITIONAL_HEADERS

    output_file = open_output_file(namespace.output)

    output_writer = csv.writer(output_file)
    output_writer.writerow(output_headers)

    loading_bar = tqdm(
        desc='Extracting content',
        total=namespace.total,
        dynamic_ncols=True,
        unit=' docs'
    )

    namespace.report.close()
    namespace.report = open(namespace.report.name)
    files = create_report_iterator(namespace, loading_bar=loading_bar)

    with Pool(namespace.processes) as pool:
        for error, line, content in pool.imap_unordered(worker, files):
            loading_bar.update()

            if error is not None:
                message = report_error(error)
                line.extend([message, ''])
                output_writer.writerow(line)
                continue

            line.extend(['', content])
            output_writer.writerow(line)

    output_file.close()



# =============================================================================
# Minet Mediacloud CLI Action
# =============================================================================
#
# Logic of the `mc` action.
#
from minet.cli.utils import die, open_output_file


def mediacloud_action(namespace):

    # A token is needed to be able to access the API
    if not namespace.token:
        die([
            'A token is needed to be able to access Mediacloud\'s API.',
            'You can provide one using the `--token` argument.'
        ])

    output_file = open_output_file(namespace.output)

    from minet.cli.mediacloud.topic import mediacloud_topic_action
    mediacloud_topic_action(namespace, output_file)

    output_file.close()



# =============================================================================
# Minet Mediacloud Topic CLI Action
# =============================================================================
#
# Logic of the `mc topic` action.
#
import csv
import json
from tqdm import tqdm
from urllib3 import Timeout

from minet.cli.mediacloud.constants import MEDIACLOUD_API_BASE_URL
from minet.utils import create_pool, request
from minet.cli.utils import print_err


def forge_url(namespace, link_id=None):
    url = '%s/topics/%s/stories/list?key=%s&limit=250' % (
        MEDIACLOUD_API_BASE_URL,
        namespace.topic_id,
        namespace.token
    )

    if link_id is not None:
        url += '&link_id=%s' % link_id

    if namespace.media_id:
        url += '&media_id=%s' % namespace.media_id

    if namespace.from_media_id:
        url += '&link_from_media_id=%s' % namespace.from_media_id

    return url


def get(http, url):
    error, response = request(http, url)

    if error:
        return error, None

    data = json.loads(response.data.decode())

    return None, data


OUPUT_HEADERS = [
    'guid',
    'stories_id',
    'title',
    'url',
    'language',
    'media_id',
    'media_name',
    'collect_date',
    'publish_date',
    'date_is_reliable',
    'facebook_share_count',
    'full_text_rss',
    'inlink_count',
    'outlink_count',
    'media_inlink_count',
    'post_count',
    'snapshots_id',
    'timespans_id',
    'next_link_id'
]


def format_csv_row(item, next_link_id):
    return [
        item['guid'],
        item['stories_id'],
        item['title'],
        item['url'],
        item['language'],
        item['media_id'],
        item['media_name'],
        item['collect_date'],
        item['publish_date'],
        item['date_is_reliable'],
        item['facebook_share_count'],
        '1' if item['full_text_rss'] else '0',
        item['inlink_count'],
        item['outlink_count'],
        item['media_inlink_count'],
        item['post_count'] or '',
        item['snapshots_id'],
        item['timespans_id'],
        next_link_id or ''
    ]


def get_next_link_id(data):

    if 'link_ids' not in data:
        return None

    pagination = data['link_ids']

    if not pagination.get('next'):
        return None

    return pagination['next']


def mediacloud_topic_action(namespace, output_file):
    link_id = None

    writer = csv.writer(output_file)
    writer.writerow(OUPUT_HEADERS)

    http = create_pool(timeout=Timeout(connect=30, read=60 * 5))

    print_err('Using the following starting url:')
    print_err(forge_url(namespace, link_id))
    print_err()

    loading_bar = tqdm(
        desc='Fetching stories',
        dynamic_ncols=True,
        unit=' stories'
    )

    while True:
        error, data = get(http, forge_url(namespace, link_id))

        if error:
            raise error

        if 'stories' not in data or len(data['stories']) == 0:
            break

        next_link_id = get_next_link_id(data)

        for story in data['stories']:
            writer.writerow(format_csv_row(story, next_link_id))

        loading_bar.update(len(data['stories']))

        if next_link_id is None:
            break

        link_id = next_link_id



# =============================================================================
# Minet CrowdTangle CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the CrowdTangle actions.
#
import csv
import json
from datetime import date, timedelta
from tqdm import tqdm
from ural import get_domain_name, normalize_url

from minet.utils import create_pool, request, RateLimiter
from minet.cli.utils import print_err, die, custom_reader
from minet.cli.crowdtangle.constants import (
    CROWDTANGLE_DEFAULT_RATE_LIMIT,
    CROWDTANGLE_DEFAULT_TIMEOUT
)

DAY_DELTA = timedelta(days=1)

URL_REPORT_HEADERS = [
    'post_ct_id',
    'post_id',
    'account_ct_id',
    'account_name',
    'account_handle',
    'original_url',
    'url',
    'normalized_url',
    'domain_name'
]


def format_url_for_csv(url_item, post):
    account = post['account']

    url = url_item['expanded']

    return [
        post['id'],
        post.get('platformId', ''),
        account['id'],
        account['name'],
        account.get('handle', ''),
        url_item['original'],
        url,
        normalize_url(url, strip_trailing_slash=True),
        get_domain_name(url)
    ]


def day_range(end):
    day_delta = timedelta(days=1)

    start_date = date(*[int(i) for i in end.split('-')])
    current_date = date.today() + day_delta

    while start_date != current_date:
        end_date = current_date
        current_date -= day_delta

        yield current_date.isoformat(), end_date.isoformat()


# TODO: __call__ should receive a status to make finer decisions
class PartitionStrategyNoop(object):
    def __init__(self, namespace, url_forge):
        self.namespace = namespace
        self.url_forge = url_forge
        self.started = False

    def __call__(self, items):
        if not self.started:
            self.started = True
            return self.url_forge(self.namespace)

        return None

    def get_postfix(self):
        return None

    def should_go_next(self, items):
        return True


class PartitionStrategyDay(PartitionStrategyNoop):
    def __init__(self, namespace, url_forge):
        self.namespace = namespace
        self.url_forge = url_forge

        self.range = day_range(namespace.start_date)

    def __call__(self, items):
        start_date, end_date = next(self.range, (None, None))

        if start_date is None:
            return None

        self.namespace.start_date = start_date
        self.namespace.end_date = end_date

        return self.url_forge(self.namespace)

    def get_postfix(self):
        return {'day': self.namespace.start_date}


class PartitionStrategyLimit(PartitionStrategyNoop):
    def __init__(self, namespace, url_forge, limit):
        self.namespace = namespace
        self.url_forge = url_forge
        self.last_item = None
        self.seen = 0
        self.shifts = 0
        self.limit = limit

    def __call__(self, items):
        if self.last_item is not None:
            if items is None:
                return None

            self.namespace.end_date = self.last_item['date'].replace(' ', 'T')

        return self.url_forge(self.namespace)

    def get_postfix(self):
        if self.namespace.end_date:
            return {'date': self.namespace.end_date, 'shifts': self.shifts}

    def should_go_next(self, items):
        n = len(items)

        if n > 0:
            self.last_item = items[-1]
            self.seen += n

        if self.seen >= self.limit:
            self.seen = 0
            self.shifts += 1
            return False

        return True


PARTITION_STRATEGIES = {
    'day': PartitionStrategyDay
}


def step(http, url, item_key):
    err, result = request(http, url)

    # Debug
    if err:
        return 'http-error', err, None

    # Bad auth
    if result.status == 401:
        return 'bad-auth', None, None

    # Bad params
    if result.status >= 400:
        return 'bad-params', result, None

    try:
        data = json.loads(result.data)['result']
    except:
        return 'bad-json', None, None

    if item_key not in data or len(data[item_key]) == 0:
        return 'exhausted', None, None

    # Extracting next link
    pagination = data['pagination']
    meta = pagination['nextPage'] if 'nextPage' in pagination else None

    return 'ok', meta, data[item_key]


def default_item_id_getter(item):
    return item['id']


def create_paginated_action(url_forge, csv_headers, csv_formatter,
                            item_name, item_key, default_rate_limit=CROWDTANGLE_DEFAULT_RATE_LIMIT,
                            item_id_getter=default_item_id_getter):

    def action(namespace, output_file):
        http = create_pool(timeout=CROWDTANGLE_DEFAULT_TIMEOUT)

        url_report_writer = None

        # Do we need to resume?
        need_to_resume = False
        if getattr(namespace, 'resume', False):
            need_to_resume = True

            if namespace.output is None:
                die(
                    'Cannot --resume without knowing the output (use -o/--output rather stdout).',
                )

            if namespace.sort_by != 'date':
                die('Cannot --resume if --sort_by is not `date`.')

            if namespace.format != 'csv':
                die('Cannot --resume jsonl format yet.')

            with open(namespace.output, 'r') as f:

                # TODO: will get ugly if file does not yet exist
                _, resume_pos, resume_reader = custom_reader(f, 'datetime')

                last_line = None
                resume_loader = tqdm(desc='Resuming', unit=' lines')

                for line in resume_reader:
                    resume_loader.update()
                    last_line = line

                resume_loader.close()

                if last_line is not None:
                    last_date = last_line[resume_pos].replace(' ', 'T')
                    namespace.end_date = last_date

                    print_err('Resuming from: %s' % last_date)

        if getattr(namespace, 'url_report', False):
            url_report_writer = csv.writer(namespace.url_report)
            url_report_writer.writerow(URL_REPORT_HEADERS)

        if getattr(namespace, 'partition_strategy', None):
            pt = namespace.partition_strategy

            if isinstance(pt, int):
                partition_strategy = PartitionStrategyLimit(namespace, url_forge, pt)
            else:

                if pt == 'day' and not namespace.start_date:
                    die('"--partition-strategy day" requires a "--start-date".')

                partition_strategy = PARTITION_STRATEGIES[pt](namespace, url_forge)
        else:
            partition_strategy = PartitionStrategyNoop(namespace, url_forge)

        N = 0
        C = 0
        url = partition_strategy(None)
        last_url = None
        last_items = set()

        has_limit = bool(namespace.limit)

        print_err('Using the following starting url:')
        print_err(url)
        print_err()

        # Loading bar
        loading_bar = tqdm(
            desc='Fetching %s' % item_name,
            dynamic_ncols=True,
            unit=' %s' % item_name,
            total=namespace.limit
        )

        def set_postfix():
            postfix = {'calls': C}

            addendum = partition_strategy.get_postfix()

            if addendum is not None:
                postfix.update(addendum)

            loading_bar.set_postfix(**postfix)

        set_postfix()

        if namespace.format == 'csv':
            writer = csv.writer(output_file)

            if not need_to_resume:
                writer.writerow(csv_headers(namespace) if callable(csv_headers) else csv_headers)

        rate_limit = namespace.rate_limit if namespace.rate_limit else default_rate_limit
        rate_limiter = RateLimiter(rate_limit, 60.0)

        # TODO: those conditions are a bit hacky. code could be clearer
        while url is not None and url != last_url:
            with rate_limiter:
                status, meta, items = step(http, url, item_key)
                C += 1

                # Debug
                if status == 'http-error':
                    loading_bar.close()
                    print_err(url)
                    raise meta

                if status == 'bad-auth':
                    loading_bar.close()
                    die([
                        'Your API token is invalid.',
                        'Check that you indicated a valid one using the `--token` argument.'
                    ])

                if status == 'bad-params':
                    loading_bar.close()
                    die([
                        'Error status %i:' % meta.status,
                        json.loads(meta.data)['message'],
                        'Last called url:',
                        url
                    ])

                if status == 'bad-json':
                    loading_bar.close()
                    die('Misformatted JSON result.')

                if status == 'exhausted':
                    url = partition_strategy(None)
                    continue

                enough_to_stop = False
                next_url = meta

                set_postfix()
                n = 0

                last_url = url

                for item in items:
                    if item_id_getter(item) in last_items:
                        continue

                    n += 1
                    N += 1

                    if namespace.format == 'jsonl':
                        output_file.write(json.dumps(item, ensure_ascii=False) + '\n')
                    else:
                        writer.writerow(csv_formatter(namespace, item))

                    if url_report_writer:
                        urls = item.get('expandedLinks')

                        if urls:
                            for url_item in urls:
                                url_report_writer.writerow(format_url_for_csv(url_item, item))

                    if has_limit and N >= namespace.limit:
                        enough_to_stop = True
                        break

                # NOTE: I wish I had labeled loops in python...
                if enough_to_stop:
                    loading_bar.close()
                    print_err('The indicated limit of %s was reached.' % item_name)
                    break
                else:
                    loading_bar.update(n)

                # We need to track last items to avoid registering the same one twice
                last_items = set(item_id_getter(item) for item in items)

                # Paginating
                if next_url is None:
                    if partition_strategy is None:
                        break

                    url = partition_strategy(items)
                    continue

                if partition_strategy.should_go_next(items):
                    url = next_url
                else:
                    url = partition_strategy(items)

        loading_bar.close()

    return action



# =============================================================================
# Minet Youtube Url Parse CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and parsing the Youtube urls
# contained in a given column.
#
from tqdm import tqdm
from ural.youtube import (
    parse_youtube_url,
    YoutubeVideo,
    YoutubeUser,
    YoutubeChannel
)

from minet.cli.utils import CSVEnricher

REPORT_HEADERS = ['youtube_type', 'youtube_id', 'youtube_name']

YOUTUBE_TYPES = {
    YoutubeVideo: 'video',
    YoutubeUser: 'user',
    YoutubeChannel: 'channel'
}


def url_parse_action(namespace, output_file):
    enricher = CSVEnricher(
        namespace.file,
        namespace.column,
        output_file,
        report_headers=REPORT_HEADERS,
        select=namespace.select.split(',') if namespace.select else None
    )

    loading_bar = tqdm(
        desc='Parsing',
        dynamic_ncols=True,
        unit=' lines',
    )

    for line in enricher:

        loading_bar.update()

        url_data = line[enricher.pos].strip()
        youtube_url = parse_youtube_url(url_data)
        if not youtube_url:
            enricher.write_empty(line)
            continue

        enricher.write(
            line,
            [YOUTUBE_TYPES.get(type(youtube_url)), youtube_url.id, getattr(youtube_url, 'name', None)]
        )



import re
from textwrap import dedent

from minet.cli.__main__ import build_parser, get_subparser
from minet.cli.commands import MINET_COMMANDS

TEMPLATE_RE = re.compile(r'<%\s+([A-Za-z/\-]+)\s+%>')


def replacer(match):
    keys = match.group(1).split('/')

    target = get_subparser(subparser_index, keys)

    return (dedent('''
        ```
        %s
        ```
    ''') % target.format_help()).strip()

parser, subparser_index = build_parser(MINET_COMMANDS)

with open('./README.template.md') as f:
    template_string = f.read()

template_string = re.sub(TEMPLATE_RE, replacer, template_string)

print(template_string)



# =============================================================================
# Minet Youtube Videos CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving metadata about
# the given Youtube videos using Google's APIs.
#
from tqdm import tqdm

from minet.cli.utils import CSVEnricher

REPORT_HEADERS = []


def videos_action(namespace, output_file):
    enricher = CSVEnricher(
        namespace.file,
        namespace.column,
        output_file,
        report_headers=REPORT_HEADERS,
        select=namespace.select.split(',') if namespace.select else None
    )

    loading_bar = tqdm(
        desc='Retrieving',
        dynamic_ncols=True,
        unit=' videos',
    )

    for line in enricher:

        loading_bar.update()

        enricher.write(
            line,
            []
        )



# =============================================================================
# Minet Facebook CLI Action Utils
# =============================================================================
#
# Miscellaneous helpers used by `minet fb`.
#
from http.cookies import SimpleCookie

from minet.utils import grab_cookies
from minet.cli.utils import die

FACEBOOK_URL = 'https://www.facebook.com/'


def fix_cookie(cookie_string):
    cookie = SimpleCookie()
    cookie.load(cookie_string)

    # NOTE: those cookie items can rat you out
    try:
        del cookie['m_pixel_ratio']
        del cookie['wd']
    except KeyError:
        pass

    cookie['locale'] = 'en_US'

    return '; '.join(key + '=' + morsel.coded_value for key, morsel in cookie.items())


def grab_facebook_cookie(namespace):
    if namespace.cookie == 'firefox' or namespace.cookie == 'chrome':
        get_cookie_for_url = grab_cookies(namespace.cookie)

        if get_cookie_for_url is None:
            die('Could not extract cookies from %s.' % namespace.cookie)

        cookie = get_cookie_for_url(FACEBOOK_URL)

    else:
        cookie = namespace.cookie.strip()

    if not cookie:
        die([
            'Relevant cookie not found.',
            'A Facebook authentication cookie is necessary to be able to access Facebook pages.',
            'Use the --cookie flag to choose a browser from which to extract the cookie or give your cookie directly.'
        ])

    return fix_cookie(cookie)



# =============================================================================
# Minet CLI Reporters
# =============================================================================
#
# Various reporters whose goal is to convert errors etc. into human-actionable
# labels in CSV format, for instance.
#
from urllib3.exceptions import (
    ConnectTimeoutError,
    MaxRetryError,
    ReadTimeoutError,
    ResponseError
)

from minet.exceptions import (
    UnknownEncodingError,
    InvalidURLError
)


def max_retry_error_reporter(error):
    if isinstance(error, ConnectTimeoutError):
        return 'connect-timeout'

    if isinstance(error, ReadTimeoutError):
        return 'read-timeout'

    if isinstance(error.reason, ResponseError) and 'redirect' in repr(error.reason):
        return 'too-many-redirects'

    return 'max-retries-exceeded'


ERROR_REPORTERS = {
    UnicodeDecodeError: 'wrong-encoding',
    UnknownEncodingError: 'unknown-encoding',
    MaxRetryError: max_retry_error_reporter,
    InvalidURLError: 'invalid-url'
}


def report_error(error):
    reporter = ERROR_REPORTERS.get(type(error), repr)

    return reporter(error) if callable(reporter) else reporter



# =============================================================================
# Minet Url Join CLI Action
# =============================================================================
#
# Logic of the `url-join` action.
#
import csv
from ural.lru import NormalizedLRUTrie
from tqdm import tqdm

from minet.cli.utils import custom_reader, open_output_file


def url_join_action(namespace):

    left_headers, left_pos, left_reader = custom_reader(namespace.file1, namespace.column1)
    right_headers, right_pos, right_reader = custom_reader(namespace.file2, namespace.column2)

    if namespace.select:
        left_headers = namespace.select.split(',')
        selected_pos = [left_headers.index(h) for h in left_headers]

    empty = [''] * len(left_headers)

    output_file = open_output_file(namespace.output)

    output_writer = csv.writer(output_file)
    output_writer.writerow(right_headers + left_headers)

    loading_bar = tqdm(
        desc='Indexing left file',
        dynamic_ncols=True,
        unit=' lines'
    )

    # First step is to index left file
    trie = NormalizedLRUTrie(strip_trailing_slash=True)

    for line in left_reader:
        url = line[left_pos].strip()

        if namespace.select:
            line = [line[p] for p in selected_pos]

        trie.set(url, line)

        loading_bar.update()

    loading_bar.close()

    loading_bar = tqdm(
        desc='Matching right file',
        dynamic_ncols=True,
        unit=' lines'
    )

    for line in right_reader:
        url = line[right_pos].strip()

        match = None

        if url:
            match = trie.match(url)

        loading_bar.update()

        if match is None:
            output_writer.writerow(line + empty)
            continue

        line.extend(match)
        output_writer.writerow(line)

    output_file.close()
